from __future__ import annotations

import itertools
import operator
from collections.abc import (
    Container,
    Sequence,
    MutableSequence,
    Mapping,
    Callable,
    Iterable,
    Iterator,
    Hashable,
)
from typing import Optional, Any, SupportsIndex, cast

from ._types import T_weight


Vector = Sequence[int]  # api.rst: documented manually
Vectors = Sequence[Vector]
Limits = Sequence[tuple[int, int]]  # api.rst: documented manually


class Position(tuple[int]):
    """A position in an n-dimensional array. It is initialized
    by a `Vector`."""

    @classmethod
    def at(cls, *coordinates: int) -> Position:
        """Factory method, that creates a position from coordinates given
        as separated parameters.
        """
        return Position(coordinates)

    def __add__(self, other: Vector) -> Position:  # type: ignore[override]
        """Add the *other* `Vector` to the position"""
        return Position(map(sum, zip(self, other)))

    def __sub__(self, other: Vector) -> Position:
        """Subtract the *other* `Vector` from the position."""
        return Position(map(operator.sub, self, other))

    def __mul__(self, multiple: SupportsIndex) -> Position:
        """Multiply each coordinate by the multiple, typically an integer.

        Attention: Since Position is a tuple, 3*p returns a tuple that
        repeats the coordinates of p three times, whilst p*3 means
        p.__mul__(3) and really multiplies each coordinate of p by 3.
        """
        i = int(multiple)
        return Position(coordinate * i for coordinate in self)

    def manhattan_distance(self, other: Vector) -> int:
        """Manhattan distance of the *other* `Vector` from the
        given position vector."""
        return sum(abs(coordinate_diff) for coordinate_diff in self - other)

    def is_in_cuboid(self, limits: Limits) -> bool:
        """Return true if each coordinate of the position is inside
        the respective range defined by the *limits* sequence (see `Limits`).
        """
        return all(
            low_limit <= coordinate < high_limit
            for coordinate, (low_limit, high_limit) in zip(self, limits)
        )

    def wrap_to_cuboid(self, limits: Limits) -> Position:
        """If a coordinate of the position is outside its respective
        limit range *(from, to)* of the *limits* sequence (see `Limits`),
        add or subtract the size (to - from + 1) of the limit range as often
        as necessary to correct this.
        """
        coordinates = []
        for coordinate, (low_limit, high_limit) in zip(self, limits):
            if coordinate < low_limit:
                window_size = high_limit - low_limit
                coordinate = high_limit - (high_limit - coordinate) % window_size
            if coordinate >= high_limit:
                window_size = high_limit - low_limit
                coordinate = (coordinate - low_limit) % window_size + low_limit
            coordinates.append(coordinate)
        return Position(coordinates)

    @staticmethod
    def moves(
        dimensions: int = 2,
        diagonals: bool = False,
        zero_move: bool = False,
        non_zero_counts: Optional[Container[int]] = None,
    ) -> Vectors:
        """
        Generate vectors of moves to neighbor positions in an n-dimensional
        Array, i.e., vectors with combinations of -1, 0 and 1 as coordinates,
        and return them in ascending order.

        The number of occurring non-zero coordinates per move can be configured.
        Default: Only moves with exactly one non-zero coordinate are generated.

        :param dimensions: Number of dimensions of the generated move vectors.
        :param diagonals: Add diagonal moves (moves with more than one non-zero
            coordinate).
        :param zero_move: Add the zero move *(0, ...)*.
        :param non_zero_counts: All moves are allowed that have a number of
            non-zero coordinates as specified by non_zero_counts, e.g.,
            non_zero_counts = {0, 1} equals zero_move = True. This option can
            not be combined with the options *diagonals* and *zero_move*.
        """

        if non_zero_counts is None:
            non_zero_counts = range(
                0 if zero_move else 1, 1 + (dimensions if diagonals else 1)
            )
        else:
            if zero_move or diagonals:
                raise RuntimeError("Incompatible options")

        return list(
            move
            for move in itertools.product(range(-1, 2), repeat=dimensions)
            if dimensions - move.count(0) in non_zero_counts
        )

    def neighbors(
        self,
        moves: Iterable[Vector],
        limits: Optional[Limits] = None,
        wrap: bool = False,
    ) -> Iterator[Position]:
        # noinspection PyShadowingNames
        """
        Iterate the positions that are reached by performing the given moves.
        Can limit or wrap moves at a cuboid.

        :param moves: They define what moves lead to neighbors (irrespective the
          optionally given limits).
        :param limits: If given, they define the cuboid the generated positions
          have to stay in.
        :param wrap: If True, moves to positions outside the limits cuboid are
          wrapped at this boundary. If False, such moves are ignored.
        """
        # inspection PyShadowingNames
        for move in moves:
            neighbor_position = self + move
            if limits is None:
                if wrap is True:
                    raise RuntimeError("Limits for Option wrap missing")
                yield neighbor_position
            elif wrap:
                yield neighbor_position.wrap_to_cuboid(limits)
            else:
                if neighbor_position.is_in_cuboid(limits):
                    yield neighbor_position


class Array:
    def __init__(self, nested_sequences: Sequence, dimensions: int = 2) -> None:
        """An n-dimensional array.

        Based on *nested sequences* that, up to a given number of
        *dimensions*, define its content. Data in nestings deeper than
        that is interpreted as part of the content of an array cell.

        Coordinates of a position in the array are meant in the order from "outer"
        to "inner" sequences along the nesting. Coordinates used as arguments
        for Array methods can be given as Vector, especially as Position.

        :param nested_sequences: Content of the array to be created.
        :param dimensions: Number of dimensions the array should have.
        """
        self.content = nested_sequences
        self.dimensions = dimensions

    def size(self) -> Sequence[int]:
        # noinspection PyShadowingNames
        """Calculate the size of the array per dimension."""
        # inspection PyShadowingNames
        size = []
        area = self.content
        for d in range(self.dimensions):
            size.append(len(area))
            area = area[0]
        return size

    def limits(self) -> Limits:
        """Calculate coordinate limits (see `Limits`) per dimension as tuples
        (from, to). Since the array consists of nested sequences,
        *from* is always zero and *to* equals the size of the array
        in this dimension."""
        return [(0, upper) for upper in self.size()]

    def mutable_copy(self) -> Array:
        """Create a mutable copy of the array."""

        def _writable(area: Sequence, dimensions: int) -> list:
            if dimensions > 1:
                return [_writable(sub_area, dimensions - 1) for sub_area in area]
            else:
                return list(area)

        return Array(_writable(self.content, self.dimensions), self.dimensions)

    def __getitem__(self, position: Vector) -> Any:
        """Get the content at the given position. This allows for using
        the expression syntax *array[...]*.

        :param position: The content at this position `Vector` is returned.
        """
        i = self.content
        for coordinate in position:
            i = i[coordinate]
        return i

    def __setitem__(self, position: Vector, content: Any) -> None:
        # noinspection PyShadowingNames
        """Set the content at the given position. Assumption: The nested
        sequence the array is build upon is mutable in its last dimension
        (i.e., a list). Allows for using the assignment syntax
        *array[...] = ...* .

        :param position: The content at this position `Vector` is replaced.
        :param content: This content is stored at the position.
        """
        # inspection PyShadowingNames
        field = self.content
        for coordinate in position[:-1]:
            field = field[coordinate]
        cast(MutableSequence, field)[position[-1]] = content

    def items(self) -> Iterator[tuple[Position, Any]]:
        """Iterate positions and content.

        :rtype: Iterator[tuple[Position, Any]]
        """
        # Above, the rtype is documented manually because Sphinx autodocs with
        # option autodoc_typehints = 'description' cannot correctly
        # evaluate the type parameters of tuple (whilst typing.Tuple works).

        def _items_in_dimension(
            area: Sequence, dimensions: int
        ) -> Iterator[tuple[tuple[int, ...], Any]]:
            if dimensions == 1:
                for coordinate, sub_area in enumerate(area):
                    yield (coordinate,), sub_area
            else:
                for coordinate, sub_area in enumerate(area):
                    for sub_vector, content in _items_in_dimension(
                        sub_area, dimensions - 1
                    ):
                        yield (coordinate,) + sub_vector, content

        return (
            (Position(coordinates), content)
            for coordinates, content in _items_in_dimension(
                self.content, self.dimensions
            )
        )

    def findall(self, content: Iterable[Any]) -> tuple[Position, ...]:
        """Find content in array and return found positions.

        :param content: Content to be searched.
        :rtype: tuple[Position, ...]
        """
        # For reason for manually documented rtype see method items above.

        content_set = set(content)

        def find_in_dimension(
            p_matrix: Sequence, p_dimensions: int
        ) -> Iterator[tuple[int, ...]]:
            if p_dimensions == 1:
                for coordinate, cell_content in enumerate(p_matrix):
                    if cell_content in content_set:
                        yield (coordinate,)
            else:
                for coordinate, sub_area in enumerate(p_matrix):
                    for found in find_in_dimension(sub_area, p_dimensions - 1):
                        yield (coordinate,) + found

        return tuple(
            Position(v) for v in find_in_dimension(self.content, self.dimensions)
        )

    def next_vertices_from_forbidden(
        self,
        forbidden: Iterable[Hashable],
        wrap: bool = False,
        diagonals: bool = False,
    ) -> Callable:
        # noinspection PyShadowingNames
        """Return a `NextVertices` function for traversal strategies, based on
        given choice of when positions qualify as neighbors (goals of a
        move) of a given position (options *wrap* and *diagonals*) and whether
        such a move is allowed w.r.t. the content of the array at this
        position (parameter *forbidden*).

        :param forbidden: A move to a position with this content is not allowed.
        :param wrap: Positions are wrapped at the array limits.
        :param diagonals: Diagonal moves are allowed.
        """
        # inspection PyShadowingNames
        forbidden_content = set(forbidden)
        limits = self.limits()
        moves = Position.moves(dimensions=self.dimensions, diagonals=diagonals)

        def next_vertices(position: Position, _: Any) -> Iterator[Position]:
            for neighbor in position.neighbors(moves=moves, limits=limits, wrap=wrap):
                if self[neighbor] not in forbidden_content:
                    yield neighbor

        return next_vertices

    def next_edges_from_cell_weights(
        self,
        content_to_weight: Mapping[Any, T_weight],
        wrap: bool = False,
        diagonals: bool = False,
    ) -> Callable[[Position, Any], Iterable[tuple[Position, T_weight]]]:
        # noinspection PyShadowingNames
        """Return a `NextEdges` function for traversal strategies, based on
        given choice of when positions qualify as neighbors (goals of a
        move) of a given position (options *wrap* and *diagonals*) and what
        weight such a move has w.r.t. the content of the matrix at this
        position (parameter *content_to_weight*, and no assigned weight means
        the move is impossible).

        :param content_to_weight: Returns the weight of a move to a given position.
        :param wrap: Positions are wrapped at the array limits.
        :param diagonals: Diagonal moves are allowed.

        :rtype: Callable[[Position, Any], Iterable[tuple[Position, T_weight]]]
        """
        # For reason for manually documented rtype see method items above.
        limits = self.limits()
        moves = Position.moves(dimensions=self.dimensions, diagonals=diagonals)

        def next_edges(vector: Position, _: Any) -> Iterator[tuple[Position, T_weight]]:
            for neighbor in vector.neighbors(moves=moves, limits=limits, wrap=wrap):
                weight = content_to_weight.get(self[neighbor], None)
                if weight is not None:
                    yield neighbor, weight

        return next_edges



from collections.abc import Iterable
from typing import Callable, Any

import nographs as nog  # noqa: F401 (used in doctests, undetected by flake 8)
from nographs import T, T_vertex


# --- Utility functions ---
def edges(
    next_edges_or_vertices: Callable[[T_vertex, Any], T],
    vertices: Iterable[T_vertex],
) -> dict[T_vertex, T]:
    # Test functionality of a next_edges or next_vertices function by calling
    # it for all vertices and return the relationship between inputs and
    # outputs in a dict.
    return dict((vertex, next_edges_or_vertices(vertex, None)) for vertex in vertices)


class SomeExamplesFromDocs:
    """
    >>> def spiral_graph(i, _):
    ...     j = (i + i // 6) % 6
    ...     yield i + 1, j * 2 + 1
    ...
    ...     if i % 2 == 0:
    ...         yield i + 6, 7 - j
    ...
    ...     else:
    ...         if i > 5:
    ...             yield i - 6, 1
    >>> traversal = nog.TraversalShortestPaths(spiral_graph).start_from(
    ...    0, build_paths=True)
    >>> traversal.go_to(5)
    5
    >>> traversal.distance
    24
    >>> traversal.paths[5]
    (0, 1, 2, 3, 4, 10, 16, 17, 11, 5)

    -- next_from_edge_dict for weighted edges given as dict --
    >>> edges_weighted_dict = {
    ...    'A': (('B', 1), ('C', 2)),
    ...    'B': (('C', 5), ('D', 5)),
    ...    'C': (('D', 2), ('E', 1), ('F', 5)),
    ...    'D': (('F', 2), ('G', 5)),
    ...    'E': (('F', 5), ('B', 1)),
    ...    'F': (('G', 2),)
    ... }

    >>> next_vertices = nog.adapt_edge_index(
    ...     edges_weighted_dict,add_inverted=False,attributes=True)
    >>> edges(next_vertices, edges_weighted_dict.keys()) == edges_weighted_dict
    True

    >>> next_vertices = nog.adapt_edge_index(
    ...     edges_weighted_dict,add_inverted=True,attributes=True)
    >>> edges(next_vertices, edges_weighted_dict.keys()
    ...      )  # doctest: +NORMALIZE_WHITESPACE
    {'A': [('B', 1), ('C', 2)], 'B': [('A', 1), ('C', 5), ('D', 5), ('E', 1)],
     'C': [('A', 2), ('B', 5), ('D', 2), ('E', 1), ('F', 5)],
     'D': [('B', 5), ('C', 2), ('F', 2), ('G', 5)],
     'E': [('C', 1), ('F', 5), ('B', 1)],
     'F': [('C', 5), ('D', 2), ('E', 5), ('G', 2)]}


    -- next_from_edge_dict for weighted edges given as sequence --
    >>> edges_weighted_sequence = (
    ...     ((1, 1), (2, 2)),
    ...     ((2, 5), (3, 5)),
    ...     ((3, 2), (4, 1), (5, 5)),
    ...     ((5, 2), (6, 5)),
    ...     ((5, 5), (1, 1)),
    ...     ((6, 2),)
    ...  )

    >>> next_vertices = nog.adapt_edge_index(edges_weighted_sequence,
    ...     add_inverted=False, attributes=True)
    >>> tuple(edges(next_vertices, range(len(edges_weighted_sequence))).values()
    ...      ) == edges_weighted_sequence
    True

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_weighted_sequence,add_inverted=True,attributes=True)
    >>> edges(next_vertices, range(len(edges_weighted_sequence))
    ...     )  # doctest: +NORMALIZE_WHITESPACE
    {0: [(1, 1), (2, 2)], 1: [(0, 1), (2, 5), (3, 5), (4, 1)],
     2: [(0, 2), (1, 5), (3, 2), (4, 1), (5, 5)], 3: [(1, 5), (2, 2), (5, 2),
     (6, 5)],
     4: [(2, 1), (5, 5), (1, 1)], 5: [(2, 5), (3, 2), (4, 5), (6, 2)]}


    -- next_from_edge_dict for unweighted edges given as dict --
    >>> edges_unweighted_dict = {
    ...    'A': ('B', 'C'),
    ...    'B': ('C', 'D'),
    ...    'C': ('D', 'E', 'F'),
    ...    'D': ('F', 'G'),
    ...    'E': ('F', 'B'),
    ...    'F': ('G', )
    ... }

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_dict,add_inverted=False,attributes=False)
    >>> edges(next_vertices, edges_weighted_dict.keys()) == edges_unweighted_dict
    True

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_dict,add_inverted=True,attributes=False)
    >>> edges(next_vertices, edges_weighted_dict.keys()
    ...      )  # doctest: +NORMALIZE_WHITESPACE
    {'A': ['B', 'C'], 'B': ['A', 'C', 'D', 'E'], 'C': ['A', 'B', 'D', 'E', 'F'],
     'D': ['B', 'C', 'F', 'G'], 'E': ['C', 'F', 'B'], 'F': ['C', 'D', 'E', 'G']}


    -- next_from_edge_dict for unweighted edges given as sequence --
    >>> edges_unweighted_sequence = (
    ...     (1, 2),
    ...     (2, 3),
    ...     (3, 4, 5),
    ...     (5, 6),
    ...     (5, 1),
    ...     (6,)
    ... )

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_sequence,add_inverted=False,attributes=False)
    >>> tuple(edges(next_vertices, range(len(edges_weighted_sequence))).values()
    ...     ) == edges_unweighted_sequence
    True

    >>> next_vertices = nog.adapt_edge_index(
    ...    edges_unweighted_sequence,add_inverted=True,attributes=False)
    >>> edges(next_vertices, range(len(edges_weighted_sequence))
    ...      )  # doctest: +NORMALIZE_WHITESPACE
    {0: [1, 2], 1: [0, 2, 3, 4], 2: [0, 1, 3, 4, 5], 3: [1, 2, 5, 6],
     4: [2, 5, 1], 5: [2, 3, 4, 6]}
    """



import nographs as nog  # noqa: F401 (used only by doctests)


class MethodsOfDummyCollectionClasses:
    """-- Methods of dummy collection classes. Objects of these classes are
    used as content of traversal objects as long as the true collections
    that provide state information of the traversal is not initialized yet,
    because the traversal has not been started so far.

    Check, if either a sensible exception raised when called.

    Note: All calls are illegal w.r.t. typing (only the number of parameters
    is correct): Instance methods are called like a classmethod would,
    the given argument for parameter self has the wrong type, and other
    arguments may be illegal, too, and the generic parameters are missing.
    But all this does not matter here, since the methods are to raise
    an Exceptioon or to return a constant value directly and in all cases.

    >>> # noinspection PyProtectedMember
    >>> strategies = nog._strategies


    >>> NoVisitedSet = strategies.NoVisitedSet
    >>> NoVisitedSet.__contains__(None, None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoVisitedSet.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoVisitedSet.__len__(None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoVisitedSet.discard(None, None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoVisitedSet.add(None, None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed


    >>> NoDistancesMapping = strategies.NoDistancesMapping
    >>> NoDistancesMapping.__getitem__(None, None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoDistancesMapping.__delitem__(None, None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoDistancesMapping.__iter__(None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoDistancesMapping.__len__(None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoDistancesMapping.__contains__(None, None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed

    >>> NoDistancesMapping.__setitem__(None, None, None)
    Traceback (most recent call last):
    RuntimeError: Traversal not started, no data to be accessed
    """



import itertools
from collections.abc import Iterator, Iterable

from ._types import T


# --- Solve 3.9 compatibility issue


def _pairwise(iterable: Iterable[T]) -> Iterator[tuple[T, T]]:
    """Returns an iterator of paired items, overlapping, from the original.
    On Python 3.10 and above, this is replaced by an alias for
    :func:`itertools.pairwise`.

    >>> list(_pairwise("abc"))
    [('a', 'b'), ('b', 'c')]
    """
    a, b = itertools.tee(iterable)
    next(b, None)
    yield from zip(a, b)


try:
    from itertools import pairwise as itertools_pairwise  # type: ignore[attr-defined]
except ImportError:  # pragma: no cover  # not executed under Python >=3.10
    pairwise = _pairwise
else:  # pragma: no cover  # not executed under Python <3.10

    def pairwise(iterable: Iterable[T]) -> Iterator[tuple[T, T]]:
        yield from itertools_pairwise(iterable)

    pairwise.__doc__ = _pairwise.__doc__



import nographs as nog  # noqa: F401 (used in doctests, undetected by flake 8)


class PathsHandling:
    # noinspection PyShadowingNames
    """
    - Error handling in Paths objects
      (Path detects illegal calls with None as vertex and with non-existing vertices.)
    - Iteration works for both types of paths both with vertices and with edges.

    >>> # noinspection PyProtectedMember
    >>> _paths = nog._paths

    -- Unlabeled paths --

    >>> gear = nog.GearDefault()
    >>> path_unlabeled = _paths.PathsOfUnlabeledEdges(
    ...    gear.vertex_id_to_vertex_mapping(()),
    ...    None
    ... )
    >>> path_unlabeled[None]  # Calls p.iter_vertices_to_start
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> None in path_unlabeled
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_unlabeled.iter_vertices_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_unlabeled.iter_edges_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_unlabeled.iter_edges_to_start(2))
    Traceback (most recent call last):
    RuntimeError: Paths: No path for given vertex.
    >>> path_unlabeled.append_edge(0, 0, None)
    >>> path_unlabeled.append_edge(0, 1, None)
    >>> # Calls p.iter_vertices_from_start and p.iter_vertices_to_start
    >>> path_unlabeled[1]
    (0, 1)
    >>> # Also calls p.iter_edges_to_start
    >>> tuple(path_unlabeled.iter_edges_from_start(1))
    ((0, 1),)
    >>> path_unlabeled.append_edge(1, 2, None)
    >>> path_unlabeled[2]
    (0, 1, 2)
    >>> tuple(path_unlabeled.iter_edges_from_start(2))
    ((0, 1), (1, 2))

    -- Paths (and not overridden in PathsOfUNlabeledEdges) --
    >>> path_unlabeled.iter_labeled_edges_from_start(2)
    Traceback (most recent call last):
    RuntimeError: Edges with labels needed, and Traversal needs to know about them
    >>> path_unlabeled.iter_labeled_edges_to_start(2)
    Traceback (most recent call last):
    RuntimeError: Edges with labels needed, and Traversal needs to know about them

    -- Labeled paths --

    >>> path_labeled = _paths.PathsOfLabeledEdges(
    ...    gear.vertex_id_to_vertex_mapping(()),
    ...    gear.vertex_id_to_vertex_mapping(()),
    ...    None
    ... )
    >>> path_labeled[None]
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> None in path_labeled
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_labeled.iter_vertices_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.
    >>> tuple(path_labeled.iter_edges_to_start(None))
    Traceback (most recent call last):
    RuntimeError: Paths: None instead of vertex given.

    >>> tuple(path_labeled.iter_edges_to_start(2))
    Traceback (most recent call last):
    RuntimeError: Paths: No path for given vertex.

    >>> path_labeled.append_edge(0, 0, (None, None))
    >>> path_labeled.append_edge(0, 1, (None, "labeled"))
    >>> tuple(path_labeled.iter_vertices_from_start(1))
    (0, 1)
    >>> tuple(path_labeled.iter_edges_from_start(1))
    ((0, 1),)
    >>> tuple(path_labeled[1])  # Calls iter_labeled_edges_from_start
    ((0, 1, 'labeled'),)

    >>> path_labeled.append_edge(1, 2, (None, "labeled"))
    >>> tuple(path_labeled.iter_vertices_from_start(2))
    (0, 1, 2)
    >>> tuple(path_labeled.iter_edges_from_start(2))
    ((0, 1), (1, 2))
    >>> path_labeled[2]  # Calls iter_labeled_edges_from_start
    ((0, 1, 'labeled'), (1, 2, 'labeled'))
    >>> tuple(path_labeled.iter_labeled_edges_to_start(2))
    ((1, 2, 'labeled'), (0, 1, 'labeled'))

    -- Any Path --
    >>> 2 in path_labeled
    True


    -- Unlabeled Path with sequence based predecessor--
    >>> gear = nog.GearForIntVertexIDsAndCFloats()
    >>> path = _paths.PathsOfUnlabeledEdges(
    ...    gear.vertex_id_to_vertex_mapping(()),
    ...    None
    ... )
    >>> path.append_edge(0, 0, [0])
    >>> path.append_edge(0, 1, [1])
    >>> path[1]
    (0, 1)

    -- Labeled Path with sequence based predecessor--
    >>> gear = nog.GearForIntVertexIDsAndCFloats()
    >>> path = _paths.PathsOfLabeledEdges(
    ...    gear.vertex_id_to_vertex_mapping(()),
    ...    gear.vertex_id_to_path_attributes_mapping(()),
    ...    None
    ... )
    >>> path.append_edge(0, 0, [0])
    >>> path.append_edge(0, 1, [1])
    >>> path[1]
    ((0, 1, 1),)


    -- Dummy paths --

    >>> # noinspection PyProtectedMember
    >>> paths_dummy = _paths.PathsDummy(nog.vertex_as_id)
    >>> # noinspection PyProtectedMember
    >>> paths_dummy._check_vertex(None)
    Traceback (most recent call last):
    RuntimeError: No paths available: Traversal not started or no paths requested.
    >>> # noinspection PyProtectedMember
    >>> paths_dummy.__getitem__(None)
    Traceback (most recent call last):
    RuntimeError: No paths available: Traversal not started or no paths requested.
    >>> # noinspection PyProtectedMember
    >>> predecessor = paths_dummy._predecessor
    >>> predecessor[None]
    Traceback (most recent call last):
    KeyError
    >>> del predecessor[None]
    Traceback (most recent call last):
    KeyError
    >>> predecessor[None] = None
    Traceback (most recent call last):
    RuntimeError: Cannot add a path, traversal not started or no paths requested.
    >>> tuple(iter(predecessor)), len(predecessor), None in predecessor
    ((), 0, False)
    """



import itertools
import unittest
from typing import Iterator

import nographs as nog


class TestCasePrimes(unittest.TestCase):
    def test_sorted(self) -> None:
        def next_edges_prime_search(
            i: int, t2: nog.TraversalShortestPathsInfBranchingSortedFlex[int, float]
        ) -> Iterator[nog.WeightedUnlabeledOutEdge[int, float]]:
            distance = t2.distance
            yield i + 1, (i + 1) - distance
            if i > 1:
                for i_next in itertools.count(i * i, i):
                    yield i_next, (i_next - distance) - 0.5

        t = nog.TraversalShortestPathsInfBranchingSorted(
            next_edges_prime_search
        ).start_from(1)
        primes = (i for i in t if i == t.distance)

        result50 = list(itertools.takewhile(lambda i: i <= 50, primes))
        primes50 = [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47]
        self.assertEqual(result50, primes50)


if __name__ == "__main__":
    unittest.main()



import nographs as nog  # noqa: F401 (used in doctests, undetected by flake 8)


class ProtocolAndABCNotImplementedErrors:
    """
    Abstract methods of protocols and ABCs.

    If the application calls them and ignores that they are abstract, an assertion
    is to be raised to inform the application about its mistake.
    Check, if this mechanism is correctly implemented.

    Note: The following calls are all illegal w.r.t. typing (only the number of
    parameters is correct): Instance methods are called like a classmethod would,
    the given argument for parameter self has the wrong type, and other arguments may
    be illegal, too, and the generic parameters are missing. But all this does not
    matter here, since the methods are to raise NotImplementedError directly and in
    all cases.


    >>> nog.GearWithoutDistances.vertex_id_set(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.GearWithoutDistances.vertex_id_to_vertex_mapping(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.GearWithoutDistances.vertex_id_to_path_attributes_mapping(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.GearWithoutDistances.sequence_of_vertices(None, None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.Gear.zero(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.Gear.infinity(None)
    Traceback (most recent call last):
    NotImplementedError

    >>> nog.Gear.vertex_id_to_distance_mapping(None, None)
    Traceback (most recent call last):
    NotImplementedError
    """


class SpecialFeatures:
    """Tests for features that are not covered by the behaviour tests
    >>> g = nog.GearForHashableVertexIDs[int, int, int, int](0.0, float("infinity"))
    >>> m = g.vertex_id_to_distance_mapping([])  #  Returns a DefaultdictWithNiceStr
    >>> str(m)  # Test __str__ of the DefaultdictWithNiceStr
    '{}'
    """



from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / 'README.rst').read_text(encoding='utf-8')

setup(
    name="nographs",
    version="3.3.1",
    description=("Graph analysis – the lazy (evaluation) way: Analysis on the fly, "
                 + "for graphs, that are computed and/or adapted on the fly."),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/HeWeMel/nographs",
    author="Dr. Helmut Melcher",
    author_email='HeWeMel@web.de',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    package_data={"nographs": ["py.typed"]},
    # install_requires=[],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Typing :: Typed",
    ],
    keywords=('graph,network,search,traverse,analysis,infinite,lazy,shortest,distance,'
              + 'depth,DFS,breadth,BFS,Dijkstra,topological,spanning,MST,TSP'),
    python_requires='>=3.9, <4',
    project_urls={
        'Documentation': 'https://nographs.readthedocs.io/',
        'Source': 'https://github.com/hewemel/nographs/',
        'Bug Reports': 'https://github.com/hewemel/nographs/issues',
        # 'Say Thanks!': 'http://saythanks.io/to/HeWeMel',
    },
    license_files='LICENSE',
)
